import typer

from changy import logic, utils

app = typer.Typer()


@app.command()
def create() -> None:
    with utils.exit_on_exception():
        logic.create_unreleased()


@app.command()
def approve() -> None:
    with utils.exit_on_exception():
        logic.approve_unreleased()



default_changelog_header = """
# Changelog

All notable changes to this project will be documented in this file.

The format is based on [Keep a Changelog](https://keepachangelog.com/en/1.1.0/),
and this project adheres to [Semantic Versioning](https://semver.org/spec/v2.0.0.html).

Generated by [Changy](https://github.com/Tiendil/changy/tree/main).
"""


default_change_file_template = """
No changes.
"""



import sys

import typer

from changy import logic, utils

app = typer.Typer()


@app.command()
def create(version: str) -> None:
    with utils.exit_on_exception():
        version_file = logic.create_version(version)

    sys.stdout.write(f"{version_file}\n")


@app.command()
def show(version: str) -> None:
    with utils.exit_on_exception():
        body = logic.show_version_changes(version)

    sys.stdout.write(body)



import contextlib
import sys

from changy import errors


@contextlib.contextmanager
def exit_on_exception():
    try:
        yield
    except errors.ChangyError as e:
        sys.stderr.write(f"Error: {e}\n")
        sys.exit(1)



from typing import Any


class Error(Exception):
    message: str | None = None

    def __init__(self, **kwargs: Any) -> None:
        for key, value in kwargs.items():
            setattr(self, key, value)

        # TODO: send to Sentry
        if "fingerprint" not in kwargs:
            self.fingerprint = None

        super().__init__((self.message or self.__class__.__name__).format(**kwargs))

    def __repr__(self) -> str:
        attributes = ", ".join(f"{key}={value}" for key, value in self.__dict__.items())

        return f"{self.__class__.__name__}: {attributes}"


class ChangyError(Error):
    pass


class ChangesDirDoesNotExist(ChangyError):
    message = "Changes directory ({directory}) does not exist"


class AlreadyInitialized(ChangyError):
    message = "Already initialized"


class NoApprovedChanges(ChangyError):
    message = "No approved changes found ({file}). Ensure you edited unreleased changes file and called `changy unrelease approve`."  # noqa: E501


class NoUnreleasedChanges(ChangyError):
    message = "No unreleased changes found ({file}). Ensure you finished updating the changelog and called `changy version create`."  # noqa: E501


class ApprovedChangesFileExists(ChangyError):
    message = "Approved changes file ({file}) should not exist at this point. Ensure you finished version generation by calling `changy version create`."  # noqa: E501


class VersionAlreadyExists(ChangyError):
    message = "Version {version} already exists. See {file}"


class VersionDoesNotExist(ChangyError):
    message = "Version {version} does not exist."



import datetime
import re
from pathlib import Path

import pydantic

from changy import constants as c
from changy import errors
from changy.settings import settings

VERSION_DATETIME_FORMAT = "%Y-%m-%dT%H-%M-%S"
CHANGES_FILE_REGEX = re.compile(r"(\d{4}-\d{2}-\d{2}T\d{2}-\d{2}-\d{2})_(.*)\.md")

workdir = Path.cwd()
configs_dir = workdir / settings.changelog_sources_dir

changelog_file = workdir / settings.changelog_name

header_file = configs_dir / settings.changelog_header
changes_template_file = configs_dir / settings.changes_file_template
unreleased_changes_file = configs_dir / settings.unreleased_changes_file
next_release_file = configs_dir / settings.next_release_changes_file


class Changes(pydantic.BaseModel):
    time: datetime.datetime
    version: str
    text: str
    file: Path

    @property
    def version_header(self) -> str:
        return f"{self.version} on {self.time.strftime('%Y-%m-%d')}"


def config_dir_must_exist():
    if not configs_dir.exists():
        raise errors.ChangesDirDoesNotExist(directory=configs_dir)


def load_changes() -> list[Changes]:
    changes_list = []

    for file in configs_dir.iterdir():
        match = CHANGES_FILE_REGEX.match(file.name)

        if not match:
            continue

        time, version = match.groups()
        text = file.read_text()

        changes = Changes(
            time=datetime.datetime.strptime(time, VERSION_DATETIME_FORMAT), file=file, version=version, text=text
        )

        changes_list.append(changes)

    changes_list.sort(key=lambda x: x.time, reverse=True)

    return changes_list


def init() -> None:

    if configs_dir.exists():
        raise errors.AlreadyInitialized()

    configs_dir.mkdir()
    header_file.write_text(c.default_changelog_header)
    changes_template_file.write_text(c.default_change_file_template)

    create_unreleased()


def create_unreleased() -> None:
    config_dir_must_exist()
    unreleased_changes_file.write_text(c.default_change_file_template)


def approve_unreleased() -> None:
    config_dir_must_exist()

    if not unreleased_changes_file.exists():
        errors.NoUnreleasedChanges(file=unreleased_changes_file)

    unreleased_changes_file.rename(next_release_file)


def create_version(version: str) -> Path:
    config_dir_must_exist()

    changes = load_changes()

    for change in changes:
        if change.version == version:
            raise errors.VersionAlreadyExists(file=change.file, version=version)

    time = datetime.datetime.now().strftime(VERSION_DATETIME_FORMAT)

    version_file_name = f"{time}_{version}.md"

    next_version_file = configs_dir / version_file_name

    if not next_release_file.exists():
        raise errors.NoApprovedChanges(file=next_release_file)

    next_release_file.rename(next_version_file)

    create_unreleased()

    return next_version_file


def show_version_changes(version: str) -> str:
    config_dir_must_exist()

    changes = load_changes()

    for change in changes:
        if change.version == version:
            return change.text.strip()

    raise errors.VersionDoesNotExist(version=version)


def create_changelog() -> None:
    config_dir_must_exist()

    if next_release_file.exists():
        raise errors.ApprovedChangesFileExists(file=next_release_file)

    if not unreleased_changes_file.exists():
        raise errors.NoUnreleasedChanges(file=unreleased_changes_file)

    header = header_file.read_text()

    releases = load_changes()

    content = [header]

    # add unreleased changes
    text = unreleased_changes_file.read_text()
    unreleased_text = f"## Unreleased\n\n{text.strip()}"
    content.append(unreleased_text)

    for change in releases:
        content.append(f"## {change.version_header}\n\n{change.text.strip()}")

    content = [x.strip() for x in content]

    changelog_file.write_text("\n\n".join(content))
